Unit Equation;

Interface

Uses Objects,IntvDbg,   Crt;

Type
  TOpType      = (otUnary,otBinary);
  TOperator    = (noAdd,noSubtract,noMultiply,noDivide,noModulo, { Numeric }
                  boAnd,boOr,boXor,boNot,                        { Boolean }
                  roEqual,roUnequal,roLess,roGreater,            { Relational }
                  roLessEqual,roGreaterEqual,
                  woAnd,woOr,woXor);                             { Bitwise }
  TVariable    = (vaP1X,vaP1Y,vaP2X,vaP2Y,vaP3X,vaP3Y,vaP4X,vaP4Y,
                  vaP5X,vaP5Y,vaP6X,vaP6Y,vaP7X,vaP7Y,vaP8X,vaP8Y,
                  vaReg0,vaReg1,vaReg2,vaReg3,vaReg4,vaReg5,vaReg6,vaReg7,
                  vaRegFlags);
  TItemType    = (itNumericOp,itBooleanOp,itRelationalOp,itBitwiseOp,
                  itNumeric,itBoolean,itVariable,itMemory,itOpenPar,
                  itClosePar,itMemRead,itMemWrite,itNone);

  TItemTypeSet = Set Of TItemType;
  TOperatorSet = Set Of TOperator;

  TItemValue = Record
   Case TItemType Of
     itNumericOp,
     itBooleanOp,
  itRelationalOp,
     itBitwiseOp: (Operator : TOperator);
       itNumeric: (Number   : LongInt);
       itBoolean: (Bool     : Boolean);
      itVariable: (Variable : TVariable);
        itMemory: (Memory   : Word);
       itOpenPar,
      itClosePar,
          itNone: (Dummy    : Word);
   End; { Case }

  PItem = ^TItem;
  TItem = Object(TObject)
    ItemType : TItemType;
    Value    : TItemValue;
    Constructor Init(IType: TItemType);
    Constructor InitItem(Item: PItem);
    Constructor InitOperator(IType: TItemType; IValue: TOperator);
    Constructor InitNumeric(IValue: LongInt);
    Constructor InitBoolean(IValue: Boolean);
    Constructor InitVariable(IValue: TVariable);
    Constructor InitMemory(IValue: Word);
    Constructor InitMemRead(IValue: Word);
    Constructor InitMemWrite(IValue: Word);
    Destructor  Done; Virtual;
  End;

Function FixNum(St: String): String;
Function Evaluate(Items: PCollection;
                  Const Buf: CP1600Mem;
                  Const Regs: CP1600Regs;
                  Const Flags: Word;
                  Const SpriteX,SpriteY: TSpriteLoc;
                  Const ReadMem,WriteMem: Boolean;
                  Const RWMemLoc: Word;
                  Var Error: Boolean): Boolean;
Function PostFix(St: String): PCollection;
{Function PostFixStr(Items: PCollection): String;}

Implementation

Uses IntvMnu,HexWrite;

Const _32BIT_OPERAND = $66;

{Uses Math;}

Function FixNum(St: String): String;
Begin
  While (St <> '') And (St[1]          = ' ') Do Delete(St,1,1);
  While (St <> '') And (St[Length(St)] = ' ') Do Delete(St,Length(St),1);
  If St = '' Then St := '0';
  FixNum := St;
End; { FixNum }

Constructor TItem.Init(IType: TItemType);
Begin
  Inherited Init;
  ItemType    := IType;
  Value.Dummy := 0;
End; { TItem.Init }

Constructor TItem.InitItem(Item: PItem);
Begin
  Inherited Init;
  ItemType := Item^.ItemType;
  Value    := Item^.Value
End; { TItem.InitItem }

Constructor TItem.InitOperator(IType: TItemType; IValue: TOperator);
Begin
  Inherited Init;
  ItemType       := IType;
  Value.Operator := IValue;
End; { TItem.InitOperator }

Constructor TItem.InitNumeric(IValue: LongInt);
Begin
  Inherited Init;
  ItemType     := itNumeric;
  Value.Number := IValue;
End; { TItem.InitNumeric }

Constructor TItem.InitBoolean(IValue: Boolean);
Begin
  Inherited Init;
  ItemType   := itBoolean;
  Value.Bool := IValue;
End; { TItem.InitBoolean }

Constructor TItem.InitVariable(IValue: TVariable);
Begin
  Inherited Init;
  ItemType       := itVariable;
  Value.Variable := IValue;
End; { TItem.InitVariable }

Constructor TItem.InitMemory(IValue: Word);
Begin
  Inherited Init;
  ItemType     := itMemory;
  Value.Memory := IValue;
End; { TItem.InitMemory }

Constructor TItem.InitMemRead(IValue: Word);
Begin
  Inherited Init;
  ItemType     := itMemRead;
  Value.Memory := IValue;
End; { TItem.InitMemRead }

Constructor TItem.InitMemWrite(IValue: Word);
Begin
  Inherited Init;
  ItemType     := itMemWrite;
  Value.Memory := IValue;
End; { TItem.InitMemWrite }

Destructor TItem.Done;
Begin
  Inherited Done;
End; { TItem.Done }

Function PostFix(St: String): PCollection;
Var
  St1       : String;
  I,J,K     : Integer;
  Stack     : PCollection;
  Items     : PCollection;
  Final     : PCollection;
  Ch        : Char;
  Finished  : Boolean;
  Error     : Boolean;
  Ok        : Boolean;
  ItemType  : TItemType;
  L         : LongInt;
  Token     : Integer;
  Item      : PItem;

  Procedure PopStack;
  Var Item: PItem;
  Begin
    If Stack^.Count > 0 Then
    Begin
      Item := PItem(Stack^.At(Stack^.Count - 1));
      Final^.Insert(New(PItem,InitItem(Item)));
      Stack^.AtFree(Stack^.Count - 1);
    End;
  End; { PopStack }

  Procedure PushStack;
  Var Item: PItem;
  Begin
    If Items^.Count > 0 Then
    Begin
      Item := PItem(Items^.At(0));
      Stack^.Insert(New(PItem,InitItem(Item)));
    End;
  End; { PushStack }

  Procedure InsertParentheses(IType: TItemType; OType: TOpType;
                              Op: TOperatorSet; Bounds: TItemTypeSet);
  Var
    OpPos    : Integer;
    I,J,K    : Integer;
    Level    : Integer;
    Finished : Boolean;
    Item     : PItem;
    Item1    : PItem;

  Begin
    OpPos  := 1;
    I      := 0;
    While I < Items^.Count Do
    Begin
      Item := PItem(Items^.At(I));

      { Detect the target operator }

      If (Item^.ItemType       = IType) And
         (Item^.Value.Operator In Op)   Then
      Begin

        { If this is a binary operator, backtrack to find the first operand }

        If OType = otBinary Then
        Begin
          OpPos    := I;
          Level    := 0;
          Finished := False;
          Repeat
            If OpPos > 0 Then
            Begin
              Dec(OpPos);
              Item1 := PItem(Items^.At(OpPos));
              If Item1^.ItemType = itClosePar Then Inc(Level);
              If Item1^.ItemType = itOpenPar  Then Dec(Level);
              If (Level = 0) And
                 (Item1^.ItemType In (Bounds + [itOpenPar])) Then
               Finished := True;
            End;
          Until (OPPos <= 0) Or Finished;
          If OpPos < I Then
          Begin
            Items^.AtInsert(OpPos,New(PItem,Init(itOpenPar)));
            Inc(I);
          End;
        End
        Else
        Begin
          Items^.AtInsert(I,New(PItem,Init(itOpenPar)));
          Inc(I);
        End;

        { Move forward to find the next operand }

        Level    := 0;
        OpPos    := I;
        Finished := False;
        Repeat
          If OpPos < (Items^.Count - 1) Then
          Begin
            Inc(OpPos);
            Item1 := PItem(Items^.At(OpPos));
            If Item1^.ItemType = itOpenPar  Then Inc(Level);
            If Item1^.ItemType = itClosePar Then Dec(Level);
            If (Level = 0) And
               (Item1^.ItemType In (Bounds + [itClosePar])) Then
             Finished := True;
          End;
        Until (OpPos = Items^.Count - 1) Or Finished;
        If OpPos > I Then
         Items^.AtInsert(OpPos + 1,New(PItem,Init(itClosePar)));
        Inc(I);
      End
      Else Inc(I);
    End; { While }
  End; { InsertParentheses }

  Procedure InsNumPar(OType: TOpType; Op: TOperatorSet);
  Begin
    InsertParentheses(itNumericOp,OType,Op,[itNumeric,itVariable,itMemory]);
  End; { InsNumPar }

  Procedure InsBoolPar(OType: TOpType; Op: TOperatorSet);
  Begin
    InsertParentheses(itBooleanOp,OType,Op,[itBoolean,itMemRead,itMemWrite]);
  End; { InsBoolPar }

  Procedure InsRelBoolPar(OType: TOpType; Op: TOperatorSet);
  Begin
    InsertParentheses(itRelationalOp,OType,Op,[itBoolean,itMemRead,itMemWrite]);
  End; { InsRelBoolPar }

  Procedure InsRelNumPar(OType: TOpType; Op: TOperatorSet);
  Begin
    InsertParentheses(itRelationalOp,OType,Op,[itNumeric,itVariable,itMemory]);
  End; { InsRelNumPar }

  Procedure InsBitwPar(OType: TOpType; Op: TOperatorSet);
  Begin
    InsertParentheses(itBitwiseOp,OType,Op,[itNumeric,itVariable,itMemory]);
  End; { InsBitwPar }

Begin

  { Check for mismatched parentheses }

  I     := 1;
  J     := 0;
  Error := False;
  While (I <= Length(St)) And Not Error Do
  Begin
    If St[I] = '(' Then Inc(J);
    If St[I] = ')' Then Dec(J);
    If J < 0 Then Error := True;
    Inc(I);
  End; { While }
  If J <> 0 Then Error := True;
  If Error Then
  Begin
    PostFix := Nil;
    Exit;
  End;

  { Initialize variables }

  Stack := New(PCollection,Init(10,10));
  Items := New(PCollection,Init(10,10));

  { Remove all spaces }

  I := Pos(' ',St);
  While I <> 0 Do
  Begin
    Delete(St,I,1);
    I := Pos(' ',St);
  End; { While }

  { Convert to uppercase }

  St := UpperCase(St);

  { Tokenize the string }

  I := 1;
  ItemType := itNone;
  Error    := False;
  While (I <= Length(St)) And Not Error Do
  Begin

    WriteLn('St = ',St);
    WriteLn('I = ',I);
    WriteLn('J = ',J);

    Token := 1;       { Default token size }

    { Detect operators (do larger ones first!) }

    If (Copy(St,I,2) = '&&') Then
    Begin
      Items^.Insert(New(PItem,InitOperator(itBooleanOp,boAnd)));
      Token := 2;
    End;
    If (Copy(St,I,2) = '||') Then
    Begin
      Items^.Insert(New(PItem,InitOperator(itBooleanOp,boOr)));
      Token := 2;
    End;
    If (Copy(St,I,2) = '^^') Then
    Begin
      Items^.Insert(New(PItem,InitOperator(itBooleanOp,boXOr)));
      Token := 2;
    End;
    If (Copy(St,I,2) = '<=') Then
    Begin
      Items^.Insert(New(PItem,InitOperator(itRelationalOp,roLessEqual)));
      Token := 2;
    End;
    If (Copy(St,I,2) = '>=') Then
    Begin
      Items^.Insert(New(PItem,InitOperator(itRelationalOp,roGreaterEqual)));
      Token := 2;
    End;
    If (Copy(St,I,2) = '<>') Then
    Begin
      Items^.Insert(New(PItem,InitOperator(itRelationalOp,roUnequal)));
      Token := 2;
    End;
    If (Copy(St,I,2) = '!=') Then
    Begin
      Items^.Insert(New(PItem,InitOperator(itRelationalOp,roUnequal)));
      Token := 2;
    End;
    If (Copy(St,I,2) = '==') Then
    Begin
      Items^.Insert(New(PItem,InitOperator(itRelationalOp,roEqual)));
      Token := 2;
    End;
    If Token = 1 Then
    Begin
      If St[I] = '~' Then Items^.Insert(New(PItem,InitOperator(itBooleanOp,boNot)));
      If St[I] = '!' Then Items^.Insert(New(PItem,InitOperator(itBooleanOp,boNot)));
      If St[I] = '=' Then Items^.Insert(New(PItem,InitOperator(itRelationalOp,roEqual)));
      If St[I] = '<' Then Items^.Insert(New(PItem,InitOperator(itRelationalOp,roLess)));
      If St[I] = '>' Then Items^.Insert(New(PItem,InitOperator(itRelationalOp,roGreater)));
      If St[I] = '&' Then Items^.Insert(New(PItem,InitOperator(itBitwiseOp,woAnd)));
      If St[I] = '|' Then Items^.Insert(New(PItem,InitOperator(itBitwiseOp,woOr)));
      If St[I] = '^' Then Items^.Insert(New(PItem,InitOperator(itBitwiseOp,woXor)));
      If St[I] = '%' Then Items^.Insert(New(PItem,InitOperator(itNumericOp,noModulo)));
      If St[I] = '*' Then Items^.Insert(New(PItem,InitOperator(itNumericOp,noMultiply)));
      If St[I] = '/' Then Items^.Insert(New(PItem,InitOperator(itNumericOp,noDivide)));
      If (St[I] = '+') And (ItemType In [itNumeric,itMemory,itVariable,itClosePar]) Then
       Items^.Insert(New(PItem,InitOperator(itNumericOp,noAdd)));
      If (St[I] = '-') And (ItemType In [itNumeric,itMemory,itVariable,itClosePar]) Then
       Items^.Insert(New(PItem,InitOperator(itNumericOp,noSubtract)));

      { Detect integers (decimal and hecadecimal) }

      If (St[I] In ['0'..'9']) Or
         ((St[I] In ['-','+']) And
          Not (ItemType In [itNumeric,itMemory,itVariable,itClosePar])) Then
      Begin
        J        := I + 1;
        Ok       := True;
        Finished := False;
        While (J <= Length(St)) And Not Finished Do
        Begin
          If Not (St[J] In ['0'..'9','A'..'F'])
           Then Finished := True
           Else Inc(J);
        End; { While }

        { Detect hexadecimal numbers }

        If (J <= Length(St)) And (St[J] = 'H') Then Inc(J);

        { Get the number's value if possible }

        Token  := J - I;
        St1    := Copy(St,I,Token);
        If St1[1] = '+' Then Delete(St1,1,1);
        If St1[Length(St1)] <> 'H' Then
        Begin
          Val(St1,L,J);
          If J <> 0 Then Error := True;
        End
        Else
        Begin
          St1 := Copy(St1,1,Length(St1) - 1);  { Strip off the "h" }
          L := 0;
          If St1[1] = '-' Then Error := True Else L := ValHex(St1);
          If L < 0 Then Error := True;
        End;
        If Not Error Then Items^.Insert(New(PItem,InitNumeric(L)));
      End;

      { Detect numeric variables }

      If (Copy(St,I,3) = 'S1X') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP1X)));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S1Y') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP1Y)));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S2X') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP2X)));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S2Y') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP2Y)));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S3X') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP3X)));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S3Y') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP3Y)));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S4X') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP4X)));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S4Y') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP4Y)));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S5X') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP5X)));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S5Y') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP5Y)));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S6X') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP6X)));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S6Y') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP6Y)));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S7X') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP7X)));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S7Y') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP7Y)));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S8X') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP8X)));
        Token := 3;
      End;
      If (Copy(St,I,3) = 'S8Y') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaP8Y)));
        Token := 3;
      End;

      If (Copy(St,I,2) = 'R0') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaReg0)));
        Token := 2;
      End;
      If (Copy(St,I,2) = 'R1') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaReg1)));
        Token := 2;
      End;
      If (Copy(St,I,2) = 'R2') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaReg2)));
        Token := 2;
      End;
      If (Copy(St,I,2) = 'R3') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaReg3)));
        Token := 2;
      End;
      If (Copy(St,I,2) = 'R4') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaReg4)));
        Token := 2;
      End;
      If (Copy(St,I,2) = 'R5') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaReg5)));
        Token := 2;
      End;
      If (Copy(St,I,2) = 'R6') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaReg6)));
        Token := 2;
      End;
      If (Copy(St,I,2) = 'R7') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaReg7)));
        Token := 2;
      End;

      If (Copy(St,I,2) = 'SW') Then
      Begin
        Items^.Insert(New(PItem,InitVariable(vaRegFlags)));
        Token := 2;
      End;
      If Token = 1 Then
      Begin

        { Detect reading from a memory location }

        If Copy(St,I,2) = 'R[' Then
        Begin
          J        := I + 2;
          Ok       := True;
          Finished := False;
          While (J <= Length(St)) And Ok And Not Finished Do
          Begin
            If Not (St[J] In ['0'..'9','A'..'F',']']) Then Ok := False;
            If St[J] = ']' Then
            Begin
              Finished := True;
              If J = I + 2 Then Ok := False;  { Detect "[]" }
            End;
            Inc(J);
          End; { While }
          If Finished And Ok Then
          Begin
            Token := J - I;
            L     := ValHex(Copy(St,I + 2,Token - 3)); { Strip off brackets }
            If L >= 0
             Then Items^.Insert(New(PItem,InitMemRead(L)))
             Else Error := True;
          End
          Else Error := True;
        End;

        { Detect writing to a memory location }

        If Copy(St,I,2) = 'W[' Then
        Begin
          J        := I + 2;
          Ok       := True;
          Finished := False;
          While (J <= Length(St)) And Ok And Not Finished Do
          Begin
            If Not (St[J] In ['0'..'9','A'..'F',']']) Then Ok := False;
            If St[J] = ']' Then
            Begin
              Finished := True;
              If J = I + 2 Then Ok := False;  { Detect "[]" }
            End;
            Inc(J);
          End; { While }
          If Finished And Ok Then
          Begin
            Token := J - I;
            L     := ValHex(Copy(St,I + 2,Token - 3)); { Strip off brackets }
            If L >= 0
             Then Items^.Insert(New(PItem,InitMemWrite(L)))
             Else Error := True;
          End
          Else Error := True;
        End;

        { Detect memory references (always hexadecimal) }

        If St[I] = '[' Then
        Begin
          J        := I + 1;
          Ok       := True;
          Finished := False;
          While (J <= Length(St)) And Ok And Not Finished Do
          Begin
            If Not (St[J] In ['0'..'9','A'..'F',']']) Then Ok := False;
            If St[J] = ']' Then
            Begin
              Finished := True;
              If J = I + 1 Then Ok := False;  { Detect "[]" }
            End;
            Inc(J);
          End; { While }
          If Finished And Ok Then
          Begin
            Token := J - I;
            L     := ValHex(Copy(St,I + 1,Token - 2)); { Strip off brackets }
            If L >= 0
             Then Items^.Insert(New(PItem,InitMemory(L)))
             Else Error := True;
          End
          Else Error := True;
        End;

        { Detect parentheses }

        If St[I] = '(' Then Items^.Insert(New(PItem,Init(itOpenPar)));
        If St[I] = ')' Then Items^.Insert(New(PItem,Init(itClosePar)));
      End;
    End;

    { Remember the type of whatever we added for next time }

    If Items^.Count > 0 Then
     ItemType := PItem(Items^.At(Items^.Count - 1))^.ItemType;

    { Move to the next token }

    Inc(I,Token);
  End; { While }

  { If the string was tokenized okay, continue }

  If Not Error Then
  Begin

    { Insert multiplier operations where multiplication is implicit }

    Final    := New(PCollection,Init(10,10));
    I        := 0;
    ItemType := itNone;
    While I < Items^.Count Do
    Begin
      Item := PItem(Items^.At(I));

      { Insert multipliers where needed }

      If (Item^.ItemType In [itNumeric,itVariable,itMemory,itOpenPar]) And
         (ItemType       In [itNumeric,itVariable,itMemory,itClosePar]) Then
      Begin
        Items^.AtInsert(I,New(PItem,InitOperator(itNumericOp,noMultiply)));
        ItemType := itNumericOp;
      End
      Else ItemType := Item^.ItemType; { Remember the type for next time }
      Inc(I);
    End; { While }

    { Add sets of parentheses for operator precedence }

    InsBoolPar(otUnary,[boNot]);
    InsNumPar(otBinary,[noMultiply,noDivide,noModulo]);
    InsBoolPar(otBinary,[boAnd]);
    InsBitwPar(otBinary,[woAnd]);
    InsNumPar(otBinary,[noAdd,noSubtract]);
    InsBoolPar(otBinary,[boOr,boXor]);
    InsBitwPar(otBinary,[woOr,woXor]);
    InsRelNumPar(otBinary,[roEqual,roUnequal,roLess,roGreater]);
    InsRelBoolPar(otBinary,[roEqual,roUnequal]);
    InsRelNumPar(otBinary,[roLessEqual,roGreaterEqual]);

    { Convert to postfix }

    While Items^.Count > 0 Do
    Begin
      Ok := True;

      ItemType := PItem(Items^.At(0))^.ItemType;
      Case ItemType Of
        itClosePar: PopStack;  { Pop an operator off the stack }

       itNumericOp,
       itBooleanOp,
    itRelationalOp,
       itBitwiseOp: PushStack; { Push any operators onto the stack }

         itNumeric,
        itVariable,
          itMemory,
         itMemRead,
        itMemWrite,
         itBoolean: Final^.Insert(New(PItem,InitItem(PItem(Items^.At(0)))));
      End; { Case }

      { Remove the item from the infix list }

      Items^.AtFree(0);
    End; { While }

    { Pop any operators off the stack }

    While Stack^.Count > 0 Do PopStack;
  End
  Else Final := Nil;
  Dispose(Items,Done);
  Dispose(Stack,Done);
  PostFix := Final;
End; { PostFix }
(*
Function PostFixStr(Items: PCollection): String;
Var
  I    : Integer;
  St   : String;
  Item : PItem;

Begin
  If Items <> Nil Then
  Begin
    St := '';
    I  := 0;
    While I < Items^.Count Do
    Begin
      Item := PItem(Items^.At(I));
      Case Item^.ItemType Of
        itNumericOp,
        itBooleanOp,
        itRelationalOp,
        itBitwiseOp:
        Begin
          Case Item^.Value.Operator Of
            noAdd: St := St + '+';
       noSubtract: St := St + '-';
       noMultiply: St := St + '*';
         noDivide: St := St + '/';
         noModulo: St := St + '%';
            boAnd: St := St + 'And';
             boOr: St := St + 'Or';
            boXor: St := St + 'Xor';
            boNot: St := St + 'Not';
          roEqual: St := St + '=';
        roUnequal: St := St + '<>';
           roLess: St := St + '<';
        roGreater: St := St + '>';
      roLessEqual: St := St + '<=';
   roGreaterEqual: St := St + '>=';
            woAnd: St := St + 'And';
             woOr: St := St + 'Or';
            woXor: St := St + 'Xor';
          End; { Case }
        End;
        itNumeric: St := St + IntToStr(Item^.Value.Number);
        itBoolean: If Item^.Value.Bool Then St := St + 'TRUE' Else St := St + 'FALSE';
        itVariable:
        Begin
          Case Item^.Value.Variable Of
          vaP1X: St := St + 'P1.X';
          vaP1Y: St := St + 'P1.Y';
          vaP2X: St := St + 'P2.X';
          vaP2Y: St := St + 'P2.Y';
          vaP3X: St := St + 'P3.X';
          vaP3Y: St := St + 'P3.Y';
          vaP4X: St := St + 'P4.X';
          vaP4Y: St := St + 'P4.Y';
          vaP5X: St := St + 'P5.X';
          vaP5Y: St := St + 'P5.Y';
          vaP6X: St := St + 'P6.X';
          vaP6Y: St := St + 'P6.Y';
          vaP7X: St := St + 'P7.X';
          vaP7Y: St := St + 'P7.Y';
          vaP8X: St := St + 'P8.X';
          vaP8Y: St := St + 'P8.Y';
         vaReg0: St := St + 'Reg.0';
         vaReg1: St := St + 'Reg.1';
         vaReg2: St := St + 'Reg.2';
         vaReg3: St := St + 'Reg.3';
         vaReg4: St := St + 'Reg.4';
         vaReg5: St := St + 'Reg.5';
         vaReg6: St := St + 'Reg.6';
         vaReg7: St := St + 'Reg.7';
     vaRegFlags: St := St + 'Reg.Flags';
          End; { Case }
        End;
        itMemory: St := St + HexWord(Item^.Value.Memory);

        { These last two should never occur for a postfix expression }

        itOpenPar: St := St + '(';
       itClosePar: St := St + ')';
      End; { Case }
      St := St + ' ';
      Inc(I);
    End; { While }
  End
  Else St := 'ERROR';
  PostFixStr := St;
End; { PostFixStr }
*)
Function Evaluate(Items: PCollection;
                  Const Buf: CP1600Mem;
                  Const Regs: CP1600Regs;
                  Const Flags: Word;
                  Const SpriteX,SpriteY: TSpriteLoc;
                  Const ReadMem,WriteMem: Boolean;
                  Const RWMemLoc: Word;
                  Var Error: Boolean): Boolean;
Type
  PItemRec = ^TItemRec;
  TItemRec = Record
    ItemType : TItemType;
    Value    : TItemValue;
  End;

Var
  Func  : TItem;
  I     : Integer;
{  Stack : TCollection;}  { TCollections take WAY too long to do this }
  E1    : TItem;          { on every instruction                      }
  E2    : TItem;
  Item  : PItem;
  Count : Integer;
  Value : TItemValue;
  Stack : PItemRec;
  SPos  : PItemRec;
  SCnt  : Integer;

  Procedure PopStack(Var Item: TItem);
  Begin
    If Not Error Then
    Begin
      If SPos <> Stack Then
      Begin
        Dec(LongInt(SPos),SizeOf(TItemRec));
        Item.ItemType := SPos^.ItemType;
        Item.Value    := SPos^.Value;
        Dec(SCnt);
      End
      Else Error := True;
{
      If Stack.Count > 0 Then
      Begin
        Item.InitItem(PItem(Stack.At(Stack.Count - 1)));
        Stack.AtFree(Stack.Count - 1);
      End
      Else Error := True;
}
    End;
  End; { PopStack }

Begin
  Error := False;
  If Items <> Nil Then
  Begin
    Count := Items^.Count;

    { There should never be more than this on the stack }

    GetMem(Stack,Count * SizeOf(TItemRec));
    SPos  := Stack;
    SCnt  := 0;
{    Stack.Init(10,10);}
    I     := 0;
    Error := False;
    While (I < Count) And Not Error Do
    Begin
      Item  := PItem(Items^.At(I));
      Value := Item^.Value;
      Func.ItemType    := itNone;
      Func.Value.Dummy := 0;

      { Handle operators }

      Case Item^.ItemType Of
        itNumericOp,
        itBooleanOp,
        itRelationalOp,
        itBitwiseOp:
        Begin
          Case Value.Operator Of
            noAdd:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E1.Value.Number + E2.Value.Number;
              End
              Else Error := True;
            End;
            noSubtract:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E2.Value.Number - E1.Value.Number;
              End
              Else Error := True;
            End;
            noMultiply:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E1.Value.Number * E2.Value.Number;
              End
              Else Error := True;
            End;
            noDivide:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) And
                 (E1.Value.Number <> 0)    Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E2.Value.Number Div E1.Value.Number;
              End
              Else Error := True;
            End;
            noModulo:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) And
                 (E1.Value.Number <> 0)    Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E2.Value.Number Mod E1.Value.Number;
              End
              Else Error := True;
            End;
            boAnd:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itBoolean) And
                 (E2.ItemType = itBoolean) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := E2.Value.Bool And E1.Value.Bool;
              End
              Else Error := True;
            End;
            boOr:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itBoolean) And
                 (E2.ItemType = itBoolean) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := E2.Value.Bool Or E1.Value.Bool;
              End
              Else Error := True;
            End;
            boXor:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itBoolean) And
                 (E2.ItemType = itBoolean) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := E2.Value.Bool Xor E1.Value.Bool;
              End
              Else Error := True;
            End;
            boNot:
            Begin
              PopStack(E1);
              If E1.ItemType = itBoolean Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := Not E1.Value.Bool;
              End
              Else If E1.ItemType = itNumeric Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := Not E1.Value.Number;
              End
              Else Error := True;
            End;
            roEqual:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itBoolean) And
                 (E2.ItemType = itBoolean) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Bool = E1.Value.Bool);
              End
              Else If (E1.ItemType = itNumeric) And
                      (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Number = E1.Value.Number);
              End
              Else Error := True;
            End;
            roUnequal:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itBoolean) And
                 (E2.ItemType = itBoolean) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Bool <> E1.Value.Bool);
              End
              Else If (E1.ItemType = itNumeric) And
                      (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Number <> E1.Value.Number);
              End
              Else Error := True;
            End;
            roLess:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Number < E1.Value.Number);
              End
              Else Error := True;
            End;
            roGreater:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Number > E1.Value.Number);
              End
              Else Error := True;
            End;
            roLessEqual:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Number <= E1.Value.Number);
              End
              Else Error := True;
            End;
            roGreaterEqual:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType   := itBoolean;
                Func.Value.Bool := (E2.Value.Number >= E1.Value.Number);
              End
              Else Error := True;
            End;
            woAnd:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E2.Value.Number And E1.Value.Number;
              End
              Else Error := True;
            End;
            woOr:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E2.Value.Number Or E1.Value.Number;
              End
              Else Error := True;
            End;
            woXor:
            Begin
              PopStack(E1);
              PopStack(E2);
              If (E1.ItemType = itNumeric) And
                 (E2.ItemType = itNumeric) Then
              Begin
                Func.ItemType     := itNumeric;
                Func.Value.Number := E2.Value.Number Xor E1.Value.Number;
              End
              Else Error := True;
            End;
          Else
            Error := True;
          End; { Case }
        End;
        itNumeric,
        itBoolean: Func.InitItem(Item);
        itVariable:
        Begin
          Func.ItemType := itNumeric;
          Case Value.Variable Of
          vaP1X: Func.Value.Number := SpriteX[0];
          vaP1Y: Func.Value.Number := SpriteY[0];
          vaP2X: Func.Value.Number := SpriteX[1];
          vaP2Y: Func.Value.Number := SpriteY[1];
          vaP3X: Func.Value.Number := SpriteX[2];
          vaP3Y: Func.Value.Number := SpriteY[2];
          vaP4X: Func.Value.Number := SpriteX[3];
          vaP4Y: Func.Value.Number := SpriteY[3];
          vaP5X: Func.Value.Number := SpriteX[4];
          vaP5Y: Func.Value.Number := SpriteY[4];
          vaP6X: Func.Value.Number := SpriteX[5];
          vaP6Y: Func.Value.Number := SpriteY[5];
          vaP7X: Func.Value.Number := SpriteX[6];
          vaP7Y: Func.Value.Number := SpriteY[6];
          vaP8X: Func.Value.Number := SpriteX[7];
          vaP8Y: Func.Value.Number := SpriteY[7];
         vaReg0: Func.Value.Number := Regs[0];
         vaReg1: Func.Value.Number := Regs[1];
         vaReg2: Func.Value.Number := Regs[2];
         vaReg3: Func.Value.Number := Regs[3];
         vaReg4: Func.Value.Number := Regs[4];
         vaReg5: Func.Value.Number := Regs[5];
         vaReg6: Func.Value.Number := Regs[6];
         vaReg7: Func.Value.Number := Regs[7];
     vaRegFlags: Func.Value.Number := Flags;
          Else
            Error := True;
          End; { Case }
        End;
        itMemory:
        Begin
          Func.ItemType     := itNumeric;
          Func.Value.Number := Buf[Value.Memory Shr 12]^[Value.Memory And $FFF]
        End;
        itMemRead:
        Begin
          Func.ItemType   := itBoolean;
          Func.Value.Bool := ReadMem And (RWMemLoc = Value.Memory)
        End;
        itMemWrite:
        Begin
          Func.ItemType   := itBoolean;
          Func.Value.Bool := WriteMem And (RWMemLoc = Value.Memory)
        End;
      End; { Case }
      Inc(I);
      If Not Error Then{ Stack.Insert(New(PItem,InitItem(@Func)));}
      Begin
        If SCnt < Count Then
        Begin
          SPos^.ItemType := Func.ItemType;
          SPos^.Value    := Func.Value;
          Inc(LongInt(SPos),SizeOf(TItemRec));
          Inc(SCnt);
        End;
      End;
    End; { While }
{    Stack.Done;}
    FreeMem(Stack,Items^.Count * SizeOf(TItemRec));
{    If Func.ItemType <> itBoolean Then Error := True;}
  End
  Else Error := True;
  If Not Error Then Evaluate := Func.Value.Bool Else Evaluate := False;
End; { Evaluate }

End.
